/*
 * ==========================================================================================
 * benchmark_rt.c
 *
 * Date: 09 July 2003
 * Author: Jerome Delamarche (c) Copyright INODES.  2003  All rights reserved.
 * Version: 1.0 
 * ------------------------------------------------------------------------------------------
 *
 * This is the helper program that receives through a pipe datae from mod_benchmark.
 * It then must maintain a table of requested URL and their response time.
 *
 * It can dump its info on SIGUSR1 reception.
 *
 * Usage:
 * 	benchmark_rt dir_name msq_name pid_name dump_name user 
 *
 * 	where:
 * 		dir_name: the name of the directory where the following files are located
 * 		msg_name: name of file used to compute the msgqueue key
 * 		pid_name: name of the file where the PID must be written
 * 		dump_name: name of file where the periodic stats must be written
 * 		user: name of the effective user
 *
 * ------------------------------------------------------------------------------------------
 * CHANGES:
 *
 * ------------------------------------------------------------------------------------------
 * TODO: 
 *
 * ==========================================================================================
 */
static const char *rcsid = "$Id$";

#include <stdio.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <stdlib.h>
#include <pwd.h>
#include "interf.h"
#include "hash.h"

/*
 * Structure of data we want to store in our Hash table for 
 * further statistical processing:
 */
typedef struct {
	long	hits;		/* count of hits for this URI since the last dump */
	long	usec;		/* total sum of micro-seconds needed to process this request since the last dump */
	long    HITS;		/* cumulative count of hits - never reset */
	long    USEC;		/* cumulative sum of msecs - never reset */
} PERF_COUNTER;

/*
 * Other statistical counter which are not URL dependant:
 */
static int	gMsgCount;	/* count of messages between two statistical dump */
static int	gSysValid;	/* 1=following datae are valid */
static int	gZombie;	/* count of zombie process */
static int	gRunning;	/* count of running process */
static int	gStopped;	/* count of stopped process */
static int	gSleeping;	/* count of sleeping process */
static int	gAvgLoad;	/* average CPU load (base 1000) */
static int	gMemUse;	/* percentage of used memory (base 1000) */
static int	gSwapUse;	/* percentage of used swap (base 1000) */
static int	gSBValid;	/* 1=following datae are valid */
static int	gServerReady;	/* count of ready instances of Apache server */
static int 	gServerBusy;	/* count of busy instances of Apache server */

/*
 * Forward Declarations:
 */
static void usage(int argc, char *argv[]);
static int StatInit(int argc, char *argv[]);
static void StatEnd();
static void DumpMsg(BenchmarkMsg *m);
static void HandleMsg(BenchmarkMsg *m);
static void StatDump();
static void StatDump1();
static void StatDump2();
static void SwitchDebug();

/*
 * Global Variables:
 */
static int 	  gDebug     = 0;	/* switched by a SIGHUP signal */
static const char *gDirName  = NULL;
static const char *gProgName = NULL;
static const char *gMsgqName = NULL;
static const char *gPidName  = NULL;
static const char *gDumpName = NULL;
static int	   gMsgID    = 0;
static int 	   gMustRun  = 1;
static MYHASH 	  *gHash     = NULL;
static PERF_COUNTER gPerf;		/* global perf counter */
static time_t	   gStartTime;
static int	   gDumpAll; 

/* 
 * -----------------------------------------------------------------
 * Main function
 * -----------------------------------------------------------------
 */
int main(int argc, char *argv[])
{
	if (StatInit(argc,argv)) return -1;

	/* clean stop... */
	signal(SIGTERM,StatEnd);
	signal(SIGINT,StatEnd);
	signal(SIGQUIT,StatEnd);

	/* Dump signals: */
	signal(SIGUSR1,StatDump1);
	signal(SIGUSR2,StatDump2);

	/* Switch debug: */
	signal(SIGHUP,SwitchDebug);

	/* wait for request: */
	while (gMustRun) {
		BenchmarkMsg m;

		int nbytes = msgrcv(gMsgID,&m,sizeof(m)-sizeof(m.mtype),0,0);

		if (nbytes < 0) {
			if (errno != EINTR)
				fprintf(stderr,"%s: read error in msgqueue: %s\n",gProgName,strerror(errno));
			continue;
		}

		/* in case someone tries to make this program crash ! */
		m.url[sizeof(m.url)-1] = '\0';

		/* if (!nbytes) continue; */
			/* should add an outer loop ? */

		/* For now: dump the rcvd data: */
		if (gDebug) DumpMsg(&m);
		HandleMsg(&m);
	}

	StatEnd();

	return 0;
}

/* 
 * -----------------------------------------------------------------
 * usage: benchmark_rt dir_name msq_name pid_name dump_name user 
 * -----------------------------------------------------------------
 */
static void usage(int argc, char *argv[])
{
	fprintf(stderr,"\nUsage: %s dir_name fifo_name pid_name dump_name user\n\n",argv[0]);
	fprintf(stderr,"where:\n");
	fprintf(stderr,"\tdir_name: directory where are located the following objects\n");
	fprintf(stderr,"\tfifo_name: FIFO where the data must be read from\n");
	fprintf(stderr,"\tpid_name: filename where to write the process ID\n");
	fprintf(stderr,"\tdump_name: filename where to write the current statistics\n");
	fprintf(stderr,"\tuser: username under which the process must run\n");

	exit(1);
}

/* 
 * ------------------
 * Init: 
 * ------------------
 */
static char *FullName(const char *dir, const char *fname)
{
	int len = strlen(dir);
	char *p = malloc(len + strlen(fname) + 1 + 1);

	strcpy(p,dir);
	if (len && dir[len-1] != '/')
		strcat(p,"/");
	strcat(p,fname);

	return p;
}

static int StatInit(int argc, char *argv[])
{
	FILE *out;
	struct passwd *pw;
	key_t key;

	gProgName = argv[0];

	if (argc < 6) usage(argc,argv);
	gDirName  = argv[1];
	gMsgqName = argv[2];
	gPidName  = FullName(gDirName,argv[3]);
	gDumpName = FullName(gDirName,argv[4]);

	/* printf("fifo=%s, pid=%s, dump=%s\n",gMsqqName,gPidName,gDumpName); */

	/* Open the msg queue in read mode: */
	key = ftok(gMsgqName,1);
	gMsgID = msgget(key,1);	/* must correspond to mod_benchmark.c declaration */
	if (gMsgID < 0) {
		fprintf(stderr,"%s: Could not open msgqueue %s/%ld in read mode: %s\n",
				gProgName,gMsgqName,key,strerror(errno));
		return -1;
	}

	/* Change the real user: */
	if (!(pw = getpwnam(argv[5]))) {
		fprintf(stderr,"%s: Invalid User '%s'\n",gProgName,argv[5]);
		return -1;
	}

	setuid(pw->pw_uid);
	/* Keep the group of the calling process (Apache) */
	/* setgid(pw->pw_gid); */

	/* Check the dump file is writeable: */
	if ((out = fopen(gDumpName,"w")) == NULL) {
		fprintf(stderr,"%s: Could not open dump file '%s' in write mode: %s\n",
				gProgName,gDumpName,strerror(errno));
		return -1;
	}
	fclose(out);

	/* Write its own PID in the specified file: */
	if ((out = fopen(gPidName,"w")) == NULL) {
		fprintf(stderr,"%s: Could not write the PID in '%s' file: %s\n",
				gProgName,gPidName,strerror(errno));
		return -1;
	}

	fprintf(out,"%d\n",getpid());
	fclose(out);

	/* Allocate a hash table: */
	gHash = HashAllocate();

	memset(&gPerf,0,sizeof(gPerf));

	gStartTime = time(NULL);

	return 0;
}

/* 
 * -----------------------
 * End  the proc scanning:
 * -----------------------
 */
static void StatEnd()
{
	/* printf("StatEnd called\n"); */

	gMustRun = 0;

	if (gPidName) unlink(gPidName);
	if (gHash) HashFree(gHash);

	exit(0);
}

/* 
 * -----------------------
 * Dump the recvd msg:
 * -----------------------
 */
static void DumpMsg(BenchmarkMsg *m)
{
	printf("--------------------------------------------------\n");
	printf("Status Code: %d\n",m->status);
	printf("Srv ready  : %d\n",m->sready);
	printf("Srv busy   : %d\n",m->sbusy);
	printf("Seconds    : %d\n",m->sec);
	printf("MSeconds   : %d\n",m->usec);
	printf("URI        : %s\n",m->url);
}

/* 
 * ----------------------------------------
 * Handle the msg sent by mod_benchmark
 * ----------------------------------------
 */
#define BENCH_TIME_LIMIT	2000000000L

static void ResetOnOverflowCallback(int hashvalue, const char *key, void *data, void *opaque)
{
	PERF_COUNTER *pc = (PERF_COUNTER *)data;

	pc->HITS = pc->USEC = 0;
}

static void HandleMsg(BenchmarkMsg *m)
{
	long s;

	/* 1-We must find the matching entry in the hash array of URIs: */
	PERF_COUNTER *pc = (PERF_COUNTER *)HashGetEntryOf(gHash,m->url);
	if (!pc) {
		/* allocate a new structure: */
		pc = (PERF_COUNTER *)HashAddEntry(gHash,m->url,sizeof(PERF_COUNTER));
		memset(pc,0,sizeof(PERF_COUNTER));
	}

	/* 2-Update the statistics: */
	s = m->usec + (m->sec * 1000000);
	pc->hits++;
	pc->usec += s;
	pc->HITS++;
	pc->USEC += s;

	/* 3-Update the global stats: */
	gPerf.hits++;
	gPerf.usec += s;
	gPerf.HITS++;
	gPerf.USEC += s;	/* can overflow ! */

	/* Note: to avoid the overflow, we reset the counters when they reach the LIMIT */
	if (gPerf.usec > BENCH_TIME_LIMIT) { gPerf.hits = gPerf.usec = 0; }
	if (gPerf.USEC > BENCH_TIME_LIMIT) { gPerf.HITS = gPerf.USEC = 0; }
	if (pc->usec > BENCH_TIME_LIMIT) { pc->hits = pc->usec = 0; }
	if (pc->USEC > BENCH_TIME_LIMIT) { pc->HITS = pc->USEC = 0; }

	/* Handle the system and the ScoreBoard stats if they are available: */
	gMsgCount++;
	gSysValid 	= m->sysvalid;
	gZombie		+= m->zombie;
	gRunning	+= m->running;
	gStopped	+= m->stopped;	
	gSleeping	+= m->sleeping;
	gAvgLoad	+= m->avgload;
	gMemUse		+= m->memuse;
	gSwapUse	+= m->swapuse;
	gSBValid	= m->sbvalid;
	gServerReady	+= m->server_ready;
	gServerBusy	+= m->server_busy;
}

/*
 * ----------------------------------------
 * Dump the statistics:
 * (add a lock during the dump ?)
 * ----------------------------------------
 */
static void DumpGlobalStat(FILE *out)
{
	fprintf(out,"Statistics Start Time: %s",ctime(&gStartTime));
	gStartTime = time(NULL);
	fprintf(out,"Statistics End Time  : %s",ctime(&gStartTime));

	/* Dump the system stats if available: */
	if (gSysValid) {
		fprintf(out,"\nSystem Statistics:\n");
		fprintf(out,"Zombie Process   # : %d\n",gZombie / gMsgCount);
		fprintf(out,"Running Process  # : %d\n",gRunning / gMsgCount);
		fprintf(out,"Stopped Process  # : %d\n",gStopped / gMsgCount);
		fprintf(out,"Sleeping Process # : %d\n",gSleeping / gMsgCount);
		fprintf(out,"Average Load       : %d\n",gAvgLoad / gMsgCount);
		fprintf(out,"% Used Memory      : %d\n",gMemUse / gMsgCount);
		fprintf(out,"% Used Swap        : %d\n",gSwapUse / gMsgCount);
	}

	/* Dump the scoreboard stats if available: */
	if (gSBValid) {
		fprintf(out,"\nScoreboard Statistics:\n");
		fprintf(out,"Server Ready # : %d\n",gServerReady / gMsgCount);
		fprintf(out,"Server Busy  # : %d\n",gServerBusy / gMsgCount);
	}

	/* Dump the URL header: */
	fprintf(out,"\n%-11s %-11s %-11s %-11s %-11s %-11s\n","Hit Count","Total Time","Mean Time","TOTAL HITS","TOTAL TIME","MEAN TIME");
	fprintf(out,"%-11ld %-11ld %-11ld %-11ld %-11ld %-11ld\n",
	       gPerf.hits,gPerf.usec,(gPerf.hits)?gPerf.usec/gPerf.hits:0,
	       gPerf.HITS,gPerf.USEC,(gPerf.HITS)?gPerf.USEC/gPerf.HITS:0);
}

static void PerfCounterDumpCallback(int hashvalue, const char *key, void *data, void *opaque)
{
	FILE *out = (FILE *)opaque;
	PERF_COUNTER *pc = (PERF_COUNTER *)data;

	/* Dump only if necessary: */
	if (gDumpAll || pc->hits) {
		fprintf(out,"%-11ld %-11ld %-11ld %-11ld %-11ld %-11ld %s\n",
			pc->hits,pc->usec,(pc->hits)?(pc->usec/pc->hits):0,
			pc->HITS,pc->USEC,(pc->HITS)?(pc->USEC/pc->HITS):0,
			key);
	}

	pc->hits = pc->usec = 0;
}

/*
 * ------------------------------------------------
 * Signal Handler: dump the current statistics
 * ------------------------------------------------
 */
static void StatDump()
{
	FILE *out;

	if ((out = fopen(gDumpName,"w")) == NULL) {
		fprintf(stderr,"%s: Could not open the dump file '%s' in write mode: %s\n",
				gProgName,gDumpName,strerror(errno));
		return;
	}

	/* First and second lines: */
	DumpGlobalStat(out);

	/* Third line: */
	fprintf(out,"%-11s %-11s %-11s %-11s %-11s %-11s URI\n","Hit Count","Total Time","Mean Time","TOTAL HITS","TOTAL TIME","MEAN TIME");

	/* Other lines... */
	HashDump(gHash,PerfCounterDumpCallback,out);

	fclose(out);

	/* Reset the global stats: */
	gPerf.hits = gPerf.usec = 0;

	/* Reset the other stats: */
	gMsgCount 	= 0;
	gSysValid	= 0;
	gZombie		= 0;
	gRunning	= 0;
	gStopped	= 0;
	gSleeping	= 0;
	gAvgLoad	= 0;
	gMemUse		= 0;
	gSwapUse	= 0;
	gSBValid	= 0;
	gServerReady	= 0;
	gServerBusy	= 0;
}

/* Dump ALL the stats, even the empty URLs rows: */
static void StatDump1()
{
	gDumpAll = 1;
	StatDump();
}

/* Does not dump the empty URLs rows (can be used for huge site to minimize the dump file size): */
static void StatDump2()
{
	gDumpAll = 0;
	StatDump();
}

/*
 * ---------------------------------------------------
 * Signal handler: switch the debug mode
 * ---------------------------------------------------
 */
static void SwitchDebug()
{
	gDebug = !gDebug;
}

/* EOF */
