/*
 * ==========================================================================================
 * mod_benchmark.c
 *
 * Date: 20 Sept 2002
 * Author: Jerome Delamarche (c) Copyright INODES.  2003  All rights reserved.
 * Version: 2.0 
 *
 * ------------------------------------------------------------------------------------------
 * CHANGES:
 *
 *	11 July 2003: 	new version 2.0
 *			support of 'benchmark_rt' helper program to provide statistics
 *			in memory (new mechanism that does not use any Database)
 *			supports Big Brother integration
 *
 *	20 Sept 2002:	handling of "ben_srvport" field
 *
 * ------------------------------------------------------------------------------------------
 *
 * TODO: Handle SMP architecture
 *
 * ------------------------------------------------------------------------------------------
 *
 * Example usage:
 *
 * LoadModule benchmark_module /apache/dir/mod_benchmark.so
 * # for Apache 1.x:
 * AddModule mod_benchmark.c
 *
 * # available module variables for LogFormat are:
 * # %{BenchmarkElapsed}n 	(x.y where x is the count of seconds and y the count of 
 * #				 microseconds)
 * # %{BenchmarkScoreBoard}n 	(x.y where x is the count of apache ready servers and y the 
 * #				 count of busy apache servers)
 * # %{BenchmarkProcStat}n	(a.b.c.d.e, where a is the count of zombie process, b the 
 * #				 count of running proc, c the count of stopped proc, d the
 * #				 count of sleeping proc and e the % of CPU Load)
 * # %{BenchmarkMemStat}n	(x.y, where x is the % of memory used and y the % of used
 * #				 swap)
 * LogFormat "%t %{BenchmarkElapsed}n %{BenchmarkScoreBoard}n \"%r\" %s %B" benchmark_log
 * CustomLog logs/benchmark.log benchmark_log
 *
 * BenchmarkTrace debug
 * BenchmarkUseScoreBoard on|off
 * BenchmarkStore path_to_db_interface.so:host:db:user:pwd
 * BenchmarkWithLogInfo on|off
 * BenchmarkWithSysInfo on|off:statprogram:socketname:interval
 * BenchmarkWithStatProg on|off:statprogram:dirname:username
 * BenchmarkAlertAction proc	// will receive the URL and the elapsed time as parameters 
 *
 * <Location />
 *	BenchmarkEngine on|off
 *	BenchmarkAccept \.html$|\.php$
 * 	BenchmarkAlertLimit milliseconds
 * </Location>
 *
 * ==========================================================================================
 */
#include "common.h"

#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "scoreboard.h"

#include <sys/time.h>
#include <sys/un.h>
#include <sys/ipc.h>
#include <signal.h>
#include <unistd.h>
#include "sysstat2.h"
#include "interf.h"

#define DEBUG 0

static const char *rcsid = "$Id$";

/* 
 * private configuration: 
 */
typedef struct {
	int	isOn;		/* directive BenchmarkEngine on|off */
	int	useScoreBoard;	/* directive BenchmarkUseScoreBoard on|off */
	int	withSysInfo;	/* directive BenchmarkWithSysInfo on|off */
	int	withStatProg;	/* directive BenchmarkWithStatProg on|off */
	int	withLogInfo;	/* directive BenchmarkWithLogInfo on|off */
	pid_t	sysstat2pid;	/* PID of sysstat2, 0 if not run */
	pid_t	statpgmpid;	/* PID of benchmark_rt, 0 if not run */
	char	*socketname;	/* socket name for data exchange with sysstat2 */
	int	interval;	/* interval between two sysstat2 computation */
	char	*dirname;	/* directory for temp files shared with benchmark_rt */
	int	msgid;		/* ID of the message queue */
	int	traceDebug;	/* directive BenchmarkTrace debug */
	ap_regex_t **regAccept;	/* from BenchmarkAccept directive */
	int 	regCount;	/* count od regex directives */
	apr_dso_handle_t *hmod;		/* handle to the DB module */
	void*   (*db_init)(apr_pool_t *pool, char *srv, char *db, char *u, char *p);	/* DB module db_init() entry point */
	int	(*db_insert)(void *opaque, const char *query);				/* DB module db_insert() entry point */
	void   	(*db_close)(void *opaque);						/* DB module db_close() entry point */
	void    *db_opaque;	/* opaque descriptor returned by db_init() */
	int	oracleDB;	/* 1 if oracle is used */
	char	*alertAction;	/* directive BenchmarkAlertAction */
	int	alertLimit;	/* directive BenchmarkAlertLimit */
} bench_config;

/* Module private variables */
static int gServerLimit;
static int gThreadLimit;

/*
 * Values of ancillary files are defined here:
 */
static char *keyname  = "/tmp/benchmark.lck2";
static char *pidname  = "benchmark.pid";
static char *lckfile1 = "benchmark.lck1";
static char *lckfile2 = "benchmark.lck2";
static char *dumpname = "benchmark.dump";

/* Forward declaration... 
 */
module benchmark_module;

/*
 * Small functions that computes the msgqueue KEY:
 */
static key_t GetMSGQKey()
{
	return ftok(keyname,1);
}

/*
 * ---------------
 * Cleanup routine: 
 * ---------------
 */
#if WITH_APACHE2
static apr_status_t benchmark_childproc_cleanup(void *v)
#else
static void benchmark_childproc_cleanup(void *v)
#endif
{
	bench_config *cfg = (bench_config *)v;

#if DEBUG
	ap_log_error(APLOG_MARK,APLOG_INFO|APLOG_NOERRNO,APR_SUCCESS,NULL,"childproc_cleanup");
#endif

	/* 1-Cleanup if "benchmark_rt" launched: */
	if (cfg->withStatProg) {
		if (cfg->dirname) {
			/* Remove the ancillary objects: */
			char *curdir = getcwd(NULL,-1);
			chdir(cfg->dirname);
			unlink(lckfile1);
			unlink(lckfile2);
			chdir(curdir);
			free(curdir);

			cfg->dirname = NULL;
		}

		/* safety check to avoid killing group of processes or the init process ! */
		if ((int)cfg->statpgmpid > 0) {
			kill(cfg->statpgmpid,SIGTERM);
			cfg->statpgmpid = 0;
			msgctl(cfg->msgid,IPC_RMID,0); 
		}
	}

	/* 2-Cleanup if "sysstat2" launched:" */
	if (cfg->withSysInfo) {
		if ((int)cfg->sysstat2pid > 0) {
			kill(cfg->sysstat2pid,SIGTERM);
			cfg->sysstat2pid = 0;
		}
	}

	/* 3-Cleanup if Database used: */
	if (cfg->db_opaque) {
		cfg->db_close(cfg->db_opaque);

#if WITH_APACHE2
		/* +++++ added on the 12 Sep 2002 */
		apr_dso_unload(cfg->hmod);
#endif
		cfg->db_opaque = NULL;
	}

#if WITH_APACHE2
	return APR_SUCCESS;
#endif
}

/* 
 * ------------------
 * Directive handlers: 
 * ------------------
 */
#if WITH_APACHE2
static int benchmark_init(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s)
#endif
#if WITH_APACHE1
static void benchmark_init(server_rec *s, apr_pool_t *p)
#endif
{
	/* +++++ removed on the 13 July 2007
	if (!ap_exists_scoreboard_image()) {
		ap_log_error(APLOG_MARK,APLOG_NOERRNO|APLOG_ERR,ALE2(APR_SUCCESS)NULL,
			"Server status unavailable in inetd mode");
	}
	*/

#if WITH_APACHE2
	ap_mpm_query(AP_MPMQ_HARD_LIMIT_DAEMONS,&gServerLimit);
	ap_mpm_query(AP_MPMQ_HARD_LIMIT_THREADS,&gThreadLimit);

	return APR_SUCCESS;
#endif
}

/*
 * -----------------------------
 * Directives at Directory Level:
 * -----------------------------
 */
static const char *benchmark_engine_cmd(cmd_parms *parms, void *config, int b)
{
	bench_config *cfg = (bench_config *)config;
	cfg->isOn = b;

	return NULL;
}

static const char *benchmark_accept_cmd(cmd_parms *parms, void *config, char *expr)
{	
	char *msg = NULL;
	bench_config *cfg = (bench_config *)config;
	ap_regex_t *reg,**newreg;

	if (!(reg = ap_pregcomp(parms->pool,expr,AP_REG_EXTENDED))) 
		return "Bad RE for BenchmarkAccept directive";

	/* allocate more room : */
	newreg = (ap_regex_t **)apr_palloc(parms->pool,sizeof(ap_regex_t *) * (cfg->regCount + 1));
	if (cfg->regCount) 
		memcpy(newreg,cfg->regAccept,sizeof(ap_regex_t *) * cfg->regCount);
	newreg[cfg->regCount] = reg;
	cfg->regAccept = newreg;
	cfg->regCount++;

	return NULL;
}

static const char *benchmark_alertlimit_cmd(cmd_parms *parms, void *config, char *proc)
{
	bench_config *cfg = (bench_config *)config;

	cfg->alertLimit = atoi(proc);
	if (cfg->alertLimit <= 0)
		return "Value for BenchmarkAlertLimit must be > 0";

	return NULL;
}

/*
 * --------------------------
 * Directives at Server Level:
 * --------------------------
 */
static const char *benchmark_trace_cmd(cmd_parms *parms, void *config, char *cmd)
{
	bench_config *cfg = (bench_config *)ap_get_module_config(parms->server->module_config,&benchmark_module);

	if (strstr(cmd,"debug"))
		cfg->traceDebug = 1;

	return NULL;
}

static const char *benchmark_scoreboard_cmd(cmd_parms *parms, void *dummy, int b)
{
	bench_config *cfg = (bench_config *)ap_get_module_config(parms->server->module_config,&benchmark_module);
	cfg->useScoreBoard = b;

	return NULL;
}

static int IsTrue(const char *onoff)
{
	return (!strcasecmp(onoff,"on") || !strcmp(onoff,"1") || !strcasecmp(onoff,"true")); 
}

static const char *benchmark_statprog_cmd(cmd_parms *parms, void *dummy, char *statinfo)
{
	/* Split statinfo into 4 fields: on/off, statpgm, dirname, user */
	char *onoff,*statpgm,*dirname,*user;
	char *msg = "Bad syntax. Some fields are missing";

	bench_config *cfg = (bench_config *)ap_get_module_config(parms->server->module_config,&benchmark_module);

#if DEBUG
	ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,ALE2(APR_SUCCESS)NULL,"benchmark_statprog_cmd called");
#endif

	onoff = statinfo;
	if (!(statpgm = strchr(onoff,':'))) return msg;
	*statpgm = 0; statpgm++;

	if (!(dirname = strchr(statpgm,':'))) return msg;
	*dirname = 0; dirname++;

	if (!(user = strchr(dirname,':'))) return msg;
	*user = 0; user++;

	/* check values :*/
	if (IsTrue(onoff)) {
		int fd;
		key_t key;
		char *curdir = getcwd(NULL,-1);
		if (chdir(dirname)) {
			ap_log_error(APLOG_MARK,APLOG_ERR,ALE2(APR_SUCCESS)parms->server,
					"Could not chdir to '%s': %s",dirname,strerror(errno));
			exit(1);
		}

		cfg->withStatProg = 1;
		cfg->statpgmpid   = 0;
		cfg->dirname      = dirname;

		/* The helper pgm does not test its uniqueness.
		 * This constraint is done here. We use the "creat/link" trick for that:
		 */
		
		/* We create the temp file, it is truncated even if it exists: */
		if ((fd = creat(lckfile1,0444)) < 0) {
			ap_log_error(APLOG_MARK,APLOG_ERR,ALE2(APR_SUCCESS)parms->server,
					"Could not create the lock file '%s': %s",lckfile1,strerror(errno));
			exit(1);
		}
		close(fd);

		/* Now try to rename it: this will fail if the new name already exists !
		 * It means another process has already create the link so the helper
		 * pgm has been launched !
		 * This test is done here because we need to know the PID of the "good"
		 * helper pgm !
		 */
		if (!link(lckfile1,lckfile2)) {
			/* Create the Message queue: */
			key = GetMSGQKey();
			cfg->msgid = msgget(key,IPC_CREAT|0666);
			if (cfg->msgid < 0) {
				ap_log_error(APLOG_MARK,APLOG_ERR,ALE2(APR_SUCCESS)parms->server,
						"Could not create the msgqueue: %s",strerror(errno));
				exit(1);
			}

			/* Start the helper pgm: */
			cfg->statpgmpid = fork();
			if (cfg->statpgmpid < 0) 
				return "Could not fork the Stat program";

			if (cfg->statpgmpid == 0) {
				/* child executes the stat pgm: */
				if (execl(statpgm,statpgm,dirname,keyname,pidname,dumpname,user,NULL) < 0) 
					ap_log_error(APLOG_MARK,APLOG_ERR,ALE2(APR_SUCCESS)parms->server,
						"Could not exec() the Stat program '%s': %s",statpgm,strerror(errno));
				exit(1);
			}
		}
		else { 
			/* open the msgqueue in write mode: */
			key = GetMSGQKey();
			cfg->msgid = msgget(key,0);
#if DEBUG
			ap_log_error(APLOG_MARK,APLOG_ERR,ALE2(APR_SUCCESS)parms->server,"msgid=%d",cfg->msgid);
#endif
			if (cfg->msgid < 0) {
				ap_log_error(APLOG_MARK,APLOG_ERR,ALE2(APR_SUCCESS)parms->server,
						"Could not open the msgqueue in write mode: %s",strerror(errno));
				exit(1);
			}
		} 

		/* Register a cleanup routine: */
		apr_pool_cleanup_register(parms->pool,cfg,benchmark_childproc_cleanup,apr_pool_cleanup_null);
	}

	return NULL;
}

static const char *benchmark_sysinfo_cmd(cmd_parms *parms, void *dummy, char *sysinfo)
{
	/* Split sysinfo into 4 fields: on/off, syspgm, socketname, interval */
	char *onoff,*syspgm,*socketname,*interval;
	char *msg = "Bad syntax. Some fields are missing";

	bench_config *cfg = (bench_config *)ap_get_module_config(parms->server->module_config,&benchmark_module);

	onoff = sysinfo;
	if (!(syspgm = strchr(onoff,':'))) return msg;
	*syspgm = 0;
	syspgm++;

	if (!(socketname = strchr(syspgm,':'))) return msg;
	*socketname = 0;
	socketname++;

	if (!(interval = strchr(socketname,':'))) return msg;
	*interval = 0; 
	interval++;

	/* check values :*/
	if (IsTrue(onoff)) {
		if (atoi(interval) < 1) 
			return "Bad value for interval";

		cfg->withSysInfo = 1;
		cfg->sysstat2pid = 0;
		cfg->socketname  = socketname;
		cfg->interval    = atoi(interval);

		/* try to start the Stat program: */
		cfg->sysstat2pid = fork();
		if (cfg->sysstat2pid < 0) 
			return "Could not fork the Stat program";

		if (cfg->sysstat2pid == 0) {
			/* child executes the stat pgm: */
			if (execl(syspgm,syspgm,socketname,interval,NULL) < 0) 
				ap_log_error(APLOG_MARK,APLOG_ERR,ALE2(APR_SUCCESS)parms->server,
					"Could not exec() the Stat program");
			exit(1);
		}

		/* Register a cleanup routine: */
		apr_pool_cleanup_register(parms->pool,cfg,benchmark_childproc_cleanup,apr_pool_cleanup_null);
	}

	return NULL;
}

static const char *benchmark_loginfo_cmd(cmd_parms *parms, void *dummy, int b)
{
	bench_config *cfg = (bench_config *)ap_get_module_config(parms->server->module_config,&benchmark_module);
	cfg->withLogInfo = b;

	return NULL;
}

static const char *benchmark_alertaction_cmd(cmd_parms *parms, void *dummy, char *proc)
{
	bench_config *cfg = (bench_config *)ap_get_module_config(parms->server->module_config,&benchmark_module);
	cfg->alertAction = proc;

	return NULL;
}

static const char *benchmark_store_cmd(cmd_parms *parms, void *config, char *storedesc)
{
	/* Split storedesc into 5 fields: dbtype,server,dbname,user,pwd: */
	char *server,*dbname, *user, *pwd;
	char *msg = "Bad syntax. Some fields are missing";
	char modname[128];
	char *fullmodname;
	apr_status_t apr;

	bench_config *cfg = ap_get_module_config(parms->server->module_config,&benchmark_module);

	if (!(server = strchr(storedesc,':'))) return msg;
	*server = 0;
	server++;

	if (!(dbname = strchr(server,':'))) return msg;
	*dbname = 0; 
	dbname++;

	if (!(user = strchr(dbname,':'))) return msg;
	*user = 0;
	user++;

	if (!(pwd = strchr(user,':'))) return msg;
	*pwd = 0;
	pwd++;

	/* Must we face with Oracle ? */
	if (strstr(storedesc,"oracle")) cfg->oracleDB = 1;

	/* Load the DB interface: */
#if WITH_APACHE2
	if ((apr = apr_dso_load(&cfg->hmod,storedesc,parms->pool)) != APR_SUCCESS) {
		char buff[512];
		apr_dso_error(cfg->hmod,buff,sizeof(buff));
		ap_log_error(APLOG_MARK,APLOG_ERR,apr,parms->server,buff);
		ap_log_error(APLOG_MARK,APLOG_ERR,apr,parms->server,
			"Module %s not found",storedesc);
		printf("%s\n",dlerror());
		return "DB module not found (see error log)";
	}

	/* Get the entry-point: */
	if ((apr = apr_dso_sym((apr_dso_handle_sym_t *)&cfg->db_init,cfg->hmod,"db_init")) != APR_SUCCESS) {
		char buff[512];
		apr_dso_error(cfg->hmod,buff,sizeof(buff));
		ap_log_error(APLOG_MARK,APLOG_ERR,apr,parms->server,buff);
		return "db_init not found in DB module";
	}
	if ((apr = apr_dso_sym((apr_dso_handle_sym_t *)&cfg->db_insert,cfg->hmod,"db_insert")) != APR_SUCCESS) 
		return "db_insert not found in DB module";
	if ((apr = apr_dso_sym((apr_dso_handle_sym_t *)&cfg->db_close,cfg->hmod,"db_close")) != APR_SUCCESS) 
		return "db_close not found in DB module";
#endif

#if WITH_APACHE1
	/* Load the DB interface: */
	if (!(cfg->hmod = ap_os_dso_load(storedesc))) {
		ap_log_error(APLOG_MARK,APLOG_ERR,parms->server,
			"Module %s not found",storedesc);
		printf("%s\n",dlerror());
		return "DB module not found (see error log)";
	}

	/* Get the entry-point: */
	if (!(cfg->db_init = dlsym(cfg->hmod,"db_init"))) 
		return "db_init not found in DB module";
	if (!(cfg->db_insert = dlsym(cfg->hmod,"db_insert")))
		return "db_insert not found in DB module";
	if (!(cfg->db_close = dlsym(cfg->hmod,"db_close")))
		return "db_close not found in DB module";
#endif

	/* Call db_init() ! */
	if (!(cfg->db_opaque = cfg->db_init(parms->pool,server,dbname,user,pwd)))
		return "Database initialization failed";

	/* Register a cleanup routine: */
	apr_pool_cleanup_register(parms->pool,cfg,benchmark_childproc_cleanup,apr_pool_cleanup_null);

	return NULL;
}

/* 
 * ------------------------
 * Module Apache directives:
 * ------------------------
 */
command_rec benchmark_cmds[] = {
    AP_INIT_FLAG("BenchmarkUseScoreBoard", benchmark_scoreboard_cmd, NULL,
     RSRC_CONF, "Use Apache ScoreBoard ?"),
    AP_INIT_TAKE1("BenchmarkWithSysInfo", benchmark_sysinfo_cmd, NULL,
     RSRC_CONF, "Store the system average load and mem usage ?"),
    AP_INIT_FLAG("BenchmarkWithLogInfo", benchmark_loginfo_cmd, NULL,
     RSRC_CONF, "Generate variables for LogFormat ?"),
    AP_INIT_TAKE1("BenchmarkWithStatProg", benchmark_statprog_cmd, NULL, 
     RSRC_CONF, "Send Data to benchmark_rt"),
    AP_INIT_TAKE1("BenchmarkTrace", benchmark_trace_cmd, NULL, 
     RSRC_CONF, "Debug listed components (debug)"),
    AP_INIT_TAKE1("BenchmarkStore", benchmark_store_cmd, NULL, 
     RSRC_CONF, "Optional definition of storage target (dbtype:dbname:user:pwd)"),
    AP_INIT_TAKE1("BenchmarkAlertAction", benchmark_alertaction_cmd, NULL, 
     RSRC_CONF, "Process to call for sending alerts"),
    AP_INIT_FLAG("BenchmarkEngine", benchmark_engine_cmd, NULL, 
     ACCESS_CONF, "Start/Stop the engine"),
    AP_INIT_TAKE1("BenchmarkAccept", benchmark_accept_cmd, NULL, 
     ACCESS_CONF, "Extended RE that list the files to be analyzed"),
    AP_INIT_TAKE1("BenchmarkAlertLimit", benchmark_alertlimit_cmd, NULL, 
      ACCESS_CONF, "Limit of time (in milliseconds) a request can take before alert"),
    {NULL}
};

/*
 * --------------------------------
 * Directory Configuration Callback:
 * --------------------------------
 */
static void *create_dir_config(apr_pool_t *p, char *path)
{
	bench_config *cfg = (bench_config *)apr_pcalloc(p,sizeof(bench_config));

	memset(cfg,0,sizeof(bench_config));

	return (void *)cfg;
}

/*
 * --------------------------------
 * Directory Configuration Callback:
 * --------------------------------
 */
static void *create_server_config(apr_pool_t *p, server_rec *r)
{
	bench_config *cfg = (bench_config *)apr_pcalloc(p,sizeof(bench_config));

	memset(cfg,0,sizeof(bench_config));

	return (void *)cfg;
}

/*
 * ------------------------------
 * Start benchmarking the request:
 * ------------------------------
 */
static int benchmark_start(request_rec *r)
{
	struct timeval stamp;
	char buff[64];
	int idx;
	bench_config *scf = (bench_config *)ap_get_module_config(r->server->module_config,&benchmark_module);
	bench_config *cfg = (bench_config *)ap_get_module_config(r->per_dir_config,&benchmark_module);

	if (scf->traceDebug)
		ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,ALE2(APR_SUCCESS)r->server,"benchmark_start on URI: %s",r->uri);

	/* benchmarking on ? */
	if (!cfg->isOn) 
		return DECLINED;

	if (scf->traceDebug) 
		ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,ALE2(APR_SUCCESS)r->server,"Benchmark Engine on for URI: %s",r->uri);

	/* matches the regex ? */
	for ( idx = 0 ; idx < cfg->regCount ; idx++ ) {
		if (regexec(cfg->regAccept[idx],r->uri,0,NULL,0))
			continue;

		/* we pass info to BenchmarkStop using headers_in because of the subquery problem */
		gettimeofday(&stamp, NULL);
		sprintf(buff,"%ld.%ld",stamp.tv_sec,stamp.tv_usec);
		apr_table_set(r->headers_in, "BenchmarkStart", apr_pstrdup(r->pool,buff));

		if (scf->traceDebug) 
			ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,ALE2(APR_SUCCESS)r->server,"Benchmark started for %s",r->uri);

		return DECLINED;
	}

    	return DECLINED;
}

/*
 * ------------------------
 * Log the benchmark result:
 * ------------------------
 */
static int benchmark_stop(request_rec *r)
{
	long took_sec, took_usec;
	struct timeval stamp_begin,stamp_end;
	char buff[128];
	const char *stamp_buff;
	int server_ready = 0;
	int server_busy = 0;
	sysstat2_t as;
	bench_config *scf = (bench_config *)ap_get_module_config(r->server->module_config,&benchmark_module);
	bench_config *cfg = (bench_config *)ap_get_module_config(r->per_dir_config,&benchmark_module);
	char buftime[64];
	time_t curtime = time(NULL);
	strftime(buftime,sizeof(buftime),"%Y-%m-%d %H:%M:%S",localtime(&curtime));

	/* Patch for Oracle: must use the TO_DATE function: */
	if (scf->oracleDB) {
		char buf[32];
		strcpy(buf,buftime);
		sprintf(buftime,"TO_DATE('%s','YYYY-MM-DD HH24:MI:SS')",buf);
	}

	/* +++++ removed on the 16 Sep 2002:
	 * +++++ when rewriting occurs, start() could have been done and the
	 * +++++ translated URI, but for the original one, "isOn" is FALSE !
	if (!cfg->isOn)
		return DECLINED;
	*/

	gettimeofday(&stamp_end, NULL);

	/* benchmark_start called ? (regex...) */
	if (!(stamp_buff = apr_table_get(r->headers_in, "BenchmarkStart")))
		return DECLINED;

	if (scf->traceDebug)
		ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,ALE2(APR_SUCCESS)r->server,"BenchmarkStop");

	sscanf(stamp_buff,"%ld.%ld",&stamp_begin.tv_sec,&stamp_begin.tv_usec);

	took_sec = stamp_end.tv_sec - stamp_begin.tv_sec;
	took_usec = stamp_end.tv_usec - stamp_begin.tv_usec;
	if (took_usec < 0) {
	    --took_sec;
	    took_usec += 1000000;
	}

	/* for logging purpose... */
	if (scf->withLogInfo) {
		sprintf(buff, "%ld.%ld", took_sec, took_usec);
		apr_table_set(r->notes,"BenchmarkElapsed",buff);
	}

	/* use scoreboard ? */
	if (scf->useScoreBoard) {
		int i,j,status;
#if WITH_APACHE2
		process_score *ps_record;
#else
		short_score score_record;
#endif
		char buff[64];

		if (scf->traceDebug)
			ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,ALE2(APR_SUCCESS)r->server,"Score board used");

#if WITH_APACHE2
		for ( i = 0 ; i < gServerLimit ; i++ ) {
			ps_record = ap_get_scoreboard_process(i);
			for ( j = 0 ; j < gThreadLimit ; j++ ) {
				worker_score *ws_record = ap_get_scoreboard_worker(i,j);
				status = ws_record->status;

				if (!ps_record->quiescing && ps_record->pid) {
					if (status == SERVER_READY && ps_record->generation == ap_my_generation)
						server_ready++;
					else if (status != SERVER_DEAD &&
						 status != SERVER_STARTING &&
						 status != SERVER_IDLE_KILL)
						server_busy++;
				}
			}
		}
#endif
#if WITH_APACHE1
		ap_sync_scoreboard_image();
		if (!ap_scoreboard_image) {
			if (scf->traceDebug)
				ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,ALE2(APR_SUCCESS)r->server,
						"Score board image is NULL (run through inetd ?)");
		}
		else {
			for ( i = 0 ; i < HARD_SERVER_LIMIT ; i++ ) {
				score_record = ap_scoreboard_image->servers[i];
				status = score_record.status;
				if (status == SERVER_READY)
					server_ready++;
				else if (status != SERVER_DEAD)
					server_busy++;
			}
		}
#endif

		if (scf->withLogInfo) {
			sprintf(buff, "%ld.%ld", server_ready, server_busy);
			apr_table_set(r->notes,"BenchmarkScoreBoard",buff);
		}
	}

	/* use sysinfo ? */
	memset(&as,0,sizeof(as));
	if (scf->withSysInfo) {
		int s;

		if (scf->traceDebug)
			ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,ALE2(APR_SUCCESS)r->server,"with sys info");

		/* connect to the socket to get the system stats: */
		if ((s = socket(AF_LOCAL,SOCK_STREAM,0)) < 0) {
			ap_log_error(APLOG_MARK,APLOG_ERR,ALE2(APR_SUCCESS)r->server,"socket() failed");
		}
		else {
			struct sockaddr_un addr;
			memset(&addr,0,sizeof(addr));
			addr.sun_family = AF_LOCAL;
			strncpy(addr.sun_path,scf->socketname,sizeof(addr.sun_path) - 1);
			addr.sun_path[sizeof(addr.sun_path)-1] = '\0';

			if (connect(s,(struct sockaddr *)&addr,sizeof(addr)) < 0) {
				ap_log_error(APLOG_MARK,APLOG_ERR,ALE2(APR_SUCCESS)r->server,"connect() failed");
			}
			else {
				char *req = "?";
				send(s,req,1,0);
				if (recv(s,&as,sizeof(as),0) != sizeof(as)) {
					ap_log_error(APLOG_MARK,APLOG_ERR,ALE2(APR_SUCCESS)r->server,"recv() failed");
				}
			}

			close(s);

			if (scf->withLogInfo) {
				sprintf(buff, "%d.%d.%d.%d.%d",as.zombie,as.running,as.stopped,as.sleeping,as.avgload);
				apr_table_set(r->notes,"BenchmarkProcStat",buff);
				sprintf(buff, "%d.%d",as.memuse,as.swapuse);
				apr_table_set(r->notes,"BenchmarkMemStat",buff);
			}
		}
	}

	/* use statprog ? */
	if (scf->withStatProg) {
		/* Send the current data to the stat pgm: 
		 *
		 * The message must be sent at once, it is made up of the following fields:
		 * status_code: int
		 * took_sec: int
		 * took_usec: int
		 * server_ready: int
		 * server_busy: int
		 * url: cstring
		 */
		int nbytes;
		BenchmarkMsg m;
		m.mtype  = getpid();
		m.sready = server_ready;
		m.sbusy  = server_busy;
		m.status = r->status;
		m.sec    = took_sec;
		m.usec   = took_usec;
		strncpy(m.url,r->uri,sizeof(m.url)-1);
		m.url[sizeof(m.url)-1] = '\0';

		/* add the sysstat2 values ? */
		if (scf->withSysInfo) {
			m.sysvalid 	= 1;
			m.zombie	= as.zombie;
			m.running	= as.running;
			m.stopped	= as.stopped;
			m.sleeping	= as.sleeping;
			m.avgload	= as.avgload;
			m.memuse	= as.memuse;
			m.swapuse	= as.swapuse;
		}
		else {
			m.sysvalid = 0;
		}

		/* add the scoreboard info ? */
		if (scf->useScoreBoard) {
			m.sbvalid	= 1;
			m.server_ready  = server_ready;
			m.server_busy	= server_busy;
		}
		else {
			m.sbvalid	= 0;
		}

		/* write info in the msg queue: */
		nbytes = sizeof(m) - sizeof(m.mtype);
		if (msgsnd(scf->msgid,&m,nbytes,IPC_NOWAIT) < 0) {
			ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,ALE2(APR_SUCCESS)r->server,
					"write error in msgqueue 0x%x (%d): %s",cfg->msgid,errno,strerror(errno));
		}
	}

	/* Log in Database ? */
	scf = (bench_config *)ap_get_module_config(r->server->module_config,&benchmark_module);
	if (scf->hmod) {
		const char *query;

		unsigned short port;
		char	*quote = "'";

#if WITH_APACHE1
		port = ntohs(r->connection->local_addr.sin_port);
#endif
#if WITH_APACHE2
		/* +++++ removed on the 15 July 2003:
		 * +++++ unnecessary with Apache2 ?
		port = ntohs(r->connection->local_addr->port);
		*/
		port = r->connection->local_addr->port;
#endif

		if (scf->oracleDB) quote = "";

		query = apr_psprintf(r->pool,
				"INSERT INTO T_BENCHMARK_LOG "
				"(LOG_DATE,URL_VALUE,ARG_VALUE,FIL_VALUE,LOG_STATUS,BEN_SEC,BEN_USEC,BEN_READY,BEN_BUSY,"
				" BEN_PZOMBIE,BEN_PSTOPPED,BEN_PRUNNING,BEN_PSLEEPING,BEN_AVGLOAD,"
				" BEN_MEMUSE,BEN_SWAPUSE,BEN_CLIADDR,BEN_SRVADDR,BEN_SRVPORT) "
				"VALUES (%s%s%s,'%s','%s','%s',%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d)",
				quote,buftime,quote,(r->uri)?r->uri:"",(r->args)?r->args:"",(r->filename)?r->filename:"",
				r->status,took_sec,took_usec,server_ready,server_busy,
				as.zombie,as.stopped,as.running,as.sleeping,as.avgload,as.memuse,as.swapuse,
				ntohl(inet_addr(r->connection->remote_ip)),ntohl(inet_addr(r->connection->local_ip)),port);

		scf->db_insert(scf->db_opaque,query);
	}

	/* Alert action specified ? */
	if (scf->alertAction && cfg->alertLimit > 0 && (took_sec * 1000 + took_usec / 1000) > cfg->alertLimit) {
		char *cmd = apr_psprintf(r->pool,"%s %s %d",scf->alertAction,r->uri,(took_sec * 1000 + took_usec / 1000));
		system(cmd);

		if (scf->hmod) {
			const char *query = apr_psprintf(r->pool,
					"INSERT INTO T_BENCHMARK_ALERT "
					"(LOG_DATE,URL_VALUE,ARG_VALUE,FIL_VALUE,BEN_SEC,BEN_USEC) "
					"VALUES ('%s','%s','%s','%s',%d,%d,%d,%d)",
					buftime,(r->uri)?r->uri:"",(r->args)?r->args:"",(r->filename)?r->filename:"",took_sec,took_usec,
					ntohl(inet_addr(r->connection->remote_ip)),ntohl(inet_addr(r->connection->local_ip)));
			scf->db_insert(scf->db_opaque,query);
		}
	}

	return DECLINED;
}

#if WITH_APACHE2
/*
 * ------------------
 * Hooks Registration:
 * ------------------
 */
static void register_hooks(apr_pool_t *pool)
{
#if DEBUG
	ap_log_error(APLOG_MARK,APLOG_ERR,APR_SUCCESS,NULL,"register mod_benchmark hooks");
#endif
	ap_hook_post_config(benchmark_init,NULL,NULL,APR_HOOK_LAST);
	/*
	 * Modified by JD on the 12 Feb 2003:
	 * APR_HOOK_FIRST instead of LAST ! (fix a bug in Apache 2.0.44+)
	 */
	ap_hook_translate_name(benchmark_start,NULL,NULL,APR_HOOK_FIRST);
	ap_hook_log_transaction(benchmark_stop,NULL,NULL,APR_HOOK_FIRST);
}

/*
 * -----------------
 * Module Definition:
 * -----------------
 */
module AP_MODULE_DECLARE_DATA benchmark_module = {
    STANDARD20_MODULE_STUFF,
    create_dir_config,			/* dir config creater */
    NULL,				/* dir merger default is to override */
    create_server_config,		/* server config */
    NULL,				/* merge server config */
    benchmark_cmds,			/* command table */
    register_hooks
};
#endif

#if WITH_APACHE1
module MODULE_VAR_EXPORT benchmark_module = {
    STANDARD_MODULE_STUFF,
    benchmark_init,			/* initializer */
    create_dir_config,			/* dir config creater */
    NULL,				/* dir merger default is to override */
    create_server_config,		/* server config */
    NULL,				/* merge server config */
    benchmark_cmds,			/* command table */
    NULL,				/* handlers */
    benchmark_start,			/* filename translation */
    NULL,				/* check_user_id */
    NULL,				/* check auth */
    NULL,				/* check access */
    NULL,				/* type_checker */
    NULL,				/* fixups */
    benchmark_stop,			/* logger */
    NULL,				/* header parser */
    NULL,				/* child_init */
    NULL,				/* child_exit */
    NULL				/* post read-request */
};
#endif

/*
 * $Log:$
 */

/* EOF */
