/*
 * ==========================================================================================
 * sysstat2.c
 *
 * Date: 09 July 2003
 * Author: Jerome Delamarche (c) Copyright INODES.  2003  All rights reserved.
 * Version: 2.0 
 * ------------------------------------------------------------------------------------------
 *
 * Daemon that binds to a Unix socket and returns
 * some system statistics when required.
 *
 * Inspired from program such as "top"
 *
 * ------------------------------------------------------------------------------------------
 * CHANGES:
 *
 * ------------------------------------------------------------------------------------------
 * TODO:
 * 
 * ==========================================================================================
 */
static const char *rcsid = "$Id$";

#include <sys/types.h>
#include <sys/stat.h>
#include <asm/param.h>	/* for the HZ definition */
#include <stdio.h>
#include <sys/errno.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <dirent.h>
#include <sys/time.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/socket.h>
#include <sys/un.h>
#include "sysstat2.h"

/*
 * Forward Declarations:
 */
static int StatInit(const char *sockname);
static int GetProcCount();
static void StatCompute(int);
static int StatSend(int s);
static void StatEnd(int);
static int ScanProc(sysstat2_t *p);
static int ScanMem(sysstat2_t *m);

/*
 * Global Variables:
 */
static sysstat2_t gTheStat;
static int gSocket = -1;
static DIR *gDir;
static int gMem;
static struct timeval gOldTime;
static pid_t gMyPid;
static int gInterval;		/* in seconds */
static pthread_mutex_t gMutex1 = PTHREAD_MUTEX_INITIALIZER;
static int gFirstStat = 1;	/* first avgload is not significative */
static int gMustRun = 1;
static int gDebug = 0;
static int gProcCount = 1;	/* # of processors */

/* 
 * ----------------------------------------------------
 * usage: sysstat2 unix_socket_name interval
 * ----------------------------------------------------
 */
void usage(int argc, char *argv[])
{
	fprintf(stderr,"Usage: %s unix_socket_name interval (in seconds)\n",argv[0]);
	exit(1);
}

int main(int argc, char *argv[])
{
	if (argc < 2) 
		usage(argc,argv);

	if ((gInterval = atoi(argv[2])) < 1) {
		fprintf(stderr,"2nd arg must be an integer: %s\n",argv[2]);
		usage(argc,argv);
	}

	if (StatInit(argv[1])) return -1;

	signal(SIGTERM,StatEnd);
	signal(SIGKILL,StatEnd);
	signal(SIGINT,StatEnd);
	signal(SIGQUIT,StatEnd);

	signal(SIGALRM,StatCompute);
	alarm(gInterval);

	/* wait for request: */
	while (gMustRun) {
		fd_set readfds;
		FD_ZERO(&readfds);
		FD_SET(gSocket,&readfds);

		if (select(gSocket+1,&readfds,NULL,NULL,NULL) == 1) {
			int s;
			struct sockaddr_un cliaddr;
			socklen_t clilen;
			if ((s = accept(gSocket,(struct sockaddr *)&cliaddr,&clilen)) >= 0) {
				char req[32];	/* one char is enough... */
				if (recv(s,req,sizeof(req),0) > 0) {
					StatSend(s);
				}
			}
		}
	}

	return 0;
}

/* 
 * ------------------
 * Init: 
 * ------------------
 */
static int StatInit(const char *sockname)
{
	struct timezone timez;
	struct sockaddr_un addr;

	/* create the Unix Domain socket: */
	if ((gSocket = socket(AF_LOCAL,SOCK_STREAM,0)) < 0) {
		fprintf(stderr,"socket() failed: %s\n",strerror(errno));
		return 1;
	}

	unlink(sockname);	/* OK if this fails */

	memset(&addr,0,sizeof(addr));
	addr.sun_family = AF_LOCAL;
	strncpy(addr.sun_path,sockname,sizeof(addr.sun_path) - 1);
	addr.sun_path[sizeof(addr.sun_path)-1] = '\0';

	if (bind(gSocket,(struct sockaddr *)&addr,sizeof(addr)) < 0) {
		fprintf(stderr,"bind() failed: %s\n",strerror(errno));
		return 2;
	}

	chmod(addr.sun_path,04777);

	if (listen(gSocket,999) < 0) {
		fprintf(stderr,"listen() failed: %s\n",strerror(errno));
		return 3;
	}

	if (!(gDir = opendir("/proc"))) {
		fprintf(stderr,"opendir() failed: %s\n",strerror(errno));
		return 4;
	}

	gettimeofday(&gOldTime,&timez);
	gMyPid = getpid();

	gProcCount = sysconf(_SC_NPROCESSORS_ONLN);

	return 0;
}

/* 
 * -----------------------
 * End  the proc scanning:
 * -----------------------
 */
static void StatEnd(int dummy)
{
	if (gDebug)
		printf("StatEnd called\n");

	pthread_mutex_lock(&gMutex1);

	if (gDir) closedir(gDir);
	if (gSocket != -1) close(gSocket);

	gMustRun = 0;

	pthread_mutex_unlock(&gMutex1);
}

/* 
 * -----------------------
 * Compute the statistics:
 * -----------------------
 */
static void StatCompute(int dummy)
{
	if (gDebug)
		printf("StatCompute called:\n");

	pthread_mutex_lock(&gMutex1);

	memset(&gTheStat,0,sizeof(gTheStat));

	if (!ScanProc(&gTheStat)) {
		if (gDebug) {
			printf("avgload=%d\n",gTheStat.avgload);
			printf("z=%d, r=%d, s=%d, s=%d\n",
					gTheStat.zombie,gTheStat.running,
					gTheStat.stopped,gTheStat.sleeping);
		}
	}

	/* must be done at each time... */
	if (!(gMem = open("/proc/meminfo",O_RDONLY))) {
		fprintf(stderr,"open(/proc/meminfo) failed: %s\n",strerror(errno));
	}
	else {
		ScanMem(&gTheStat);
		close(gMem);

		if (gDebug) {
			printf("memuse=%d, swapuse=%d\n",gTheStat.memuse,gTheStat.swapuse);
		}
	}

	alarm(gInterval);

	pthread_mutex_unlock(&gMutex1);
}

/* 
 * ---------------------------
 * Send the current stats back:
 * ---------------------------
 */
static int StatSend(int s)
{
	send(s,&gTheStat,sizeof(gTheStat),0);
}

/* 
 * -----------------------
 * Scan the proc table:
 * -----------------------
 */
static int ScanProc(sysstat2_t *p)
{
	struct dirent *ent;
	struct stat sbuf;
	int fd;
	char *p1;
	int n;
	char buff[512];
	char path[64];
	char state;
	int dummy;
	long ldummy;
	long utime,stime;
	long total_time,pcpu;
	long idle_ticks = 1000;

    	/* Compute the elapsed time since last call: */
	struct timeval curtime;
	struct timezone timez;
	float elapsed_time;

	gettimeofday(&curtime, &timez);
	elapsed_time = (curtime.tv_sec - gOldTime.tv_sec) +
			(float) (curtime.tv_usec - gOldTime.tv_usec) / 1000000.0;
	gOldTime.tv_sec  = curtime.tv_sec;
	gOldTime.tv_usec = curtime.tv_usec;

	if (gDebug)
		printf("ScanProc called\n");

	/* +++++ added on the 12 July 2003: */
	rewinddir(gDir);

	while (ent = readdir(gDir)) {
		if (gDebug) 
			printf("%s ",ent->d_name);

		if (*ent->d_name < '0' || *ent->d_name > '9')
			continue;

		// ignore our own process to evaluate the global CPU load:
		if (atol(ent->d_name) == gMyPid)
			continue;

		sprintf(path,"/proc/%s",ent->d_name);
		if (stat(path,&sbuf) == -1) {
			fprintf(stderr,"ScanProc: stat(%s) failed",path);
			continue;
		}

		strcat(path,"/stat");
		if ((fd = open(path,O_RDONLY,0)) < 0) {
			fprintf(stderr,"ScanProc: open(%s) failed",path);
			continue;
		}

		if ((n = read(fd,buff,sizeof(buff)-1)) <= 0) {
			fprintf(stderr,"ScanProc: read(%s) failed",path);
			close(fd);
			continue;
		}
		buff[n] = 0;

		if (!(p1 = strrchr(buff,')'))) {
			fprintf(stderr,"ScanProc: missing ) in %s",path);
			close(fd);
			continue;
		}

		*p1 = 0;
		
		sscanf(p1+2,
			"%c %d %d %d %d %d %lu %lu %lu %lu %lu %ld %ld",
			&state, &dummy, &dummy, &dummy, &dummy, &dummy,
			&ldummy, &ldummy, &ldummy, &ldummy, &ldummy,
			&utime, &stime);

		close(fd);

		switch (state) {
		case 'S':
		case 'D': p->sleeping++; break;
		case 'Z': p->zombie++; break;
		case 'T': p->stopped++; break;
		case 'R': p->running++; break;
		}

		total_time = utime + stime;
		pcpu = (total_time * 10 * 100/HZ) / elapsed_time;
		if (pcpu > 999)
		    pcpu = 999;

		idle_ticks -= pcpu;
	}

	if (idle_ticks < 0)
		idle_ticks = 0;

	if (gFirstStat) {
		p->avgload= 0;
		gFirstStat = 0;
	}
	else {
		p->avgload = (1000 - idle_ticks) / 10;

		/* ++++++ added on the 14 Aug 2003:
		 * ++++++ SMP are taken in account by dividing the global load by the # of process:
		 */
		p->avgload /= gProcCount;
	}

	return 0;
}

/* 
 * -----------------------
 * Scan the mem info:
 * for linux version > 2
 * -----------------------
 */
static int ScanMem(sysstat2_t *m)
{
	char buff[1024],*p;
	int n;
	char fieldbuf[12];
	int k;
	unsigned long mem_total = 0,mem_free = 0;
	unsigned long swap_total = 0,swap_free = 0;

	if (gDebug)
		printf("ScanMem called\n");

	/* skip the comment line: */
	if (!(n = read(gMem,buff,sizeof(buff)-1))) {
		fprintf(stderr,"read() error on meminfo (n=%d)",n);
		return 1;
	}

	buff[n] = '\0';
	p = buff;

	/* compute the mem usage: */
	while (*p) {
	    	sscanf(p,"%11s%n",fieldbuf,&k);
	    	if(!strcmp(fieldbuf,"MemTotal:")) {
	    		p+=k;
	    		sscanf(p," %d",&mem_total);
	    		while(*p++ != '\n');
	    	}
	    	else if(!strcmp(fieldbuf,"MemFree:")) {
	    		p+=k;
	    		sscanf(p," %d",&mem_free);
	    		while(*p++ != '\n');
	    	}
    		else if(!strcmp(fieldbuf,"SwapTotal:")) {
    			p+=k;
    			sscanf(p," %d",&swap_total);
    			while(*p++ != '\n');
    		}
    		else if(!strcmp(fieldbuf,"SwapFree:")) {
    			p+=k;
    			sscanf(p," %d",&swap_free);
    			while(*p++ != '\n');
    		}
    		else
    			while(*p++ != '\n'); /* ignore lines we don't want */

		if (mem_total && mem_free && swap_total && swap_free)
			break;
    	}		

	if (mem_total && mem_free)
		m->memuse = ((mem_total - mem_free) * 100) / mem_total;
	if (swap_total && swap_free)
		m->swapuse = ((swap_total - swap_free) * 100) / swap_total;

	return 0;
}

/*
 * Finds the current time (in microseconds) and calculates the time
 * elapsed since the last update. This is essential for computing
 * percent CPU usage.
 */
static float GetElapsedTime(void)
{
    struct timeval time;
    static struct timeval oldtime;
    struct timezone timez;
    float elapsed_time;

    gettimeofday(&time, &timez);
    elapsed_time = (time.tv_sec - oldtime.tv_sec)
	+ (float) (time.tv_usec - oldtime.tv_usec) / 1000000.0;
    oldtime.tv_sec = time.tv_sec;
    oldtime.tv_usec = time.tv_usec;
    return elapsed_time;
}

/* EOF */
